from typing import Callable, Protocol, runtime_checkable

from omnipy.api.typedefs import DecoratorClassT


@runtime_checkable
class IsCallableParamAfterSelf(Protocol):
    """"""
    def __call__(self, callable_arg: Callable, /, *args: object, **kwargs: object) -> None:
        ...


@runtime_checkable
class IsCallableClass(Protocol[DecoratorClassT]):
    """"""
    def __call__(self, *args: object, **kwargs: object) -> Callable[[Callable], DecoratorClassT]:
        ...



from collections import UserDict
import json
import os
import tarfile
from typing import Annotated, Any, Generic, get_args, get_origin, Iterator, Optional, Type, TypeVar

# from orjson import orjson
from pydantic import Field, PrivateAttr, root_validator, ValidationError
from pydantic.fields import Undefined, UndefinedType
from pydantic.generics import GenericModel
from pydantic.utils import lenient_issubclass

from omnipy.data.model import _cleanup_name_qualname_and_module, Model
from omnipy.util.helpers import is_optional, is_strict_subclass, remove_forward_ref_notation

ModelT = TypeVar('ModelT', bound=Model)
DATA_KEY = 'data'

# def orjson_dumps(v, *, default):
#     # orjson.dumps returns bytes, to match standard json.dumps we need to decode
#     return orjson.dumps(v, default=default).decode()

# TODO: implement copy(), __copy__() and __deepcopy__() for Dataset and Model, making use of
#       BaseModel.copy()


class Dataset(GenericModel, Generic[ModelT], UserDict):
    """
    Dict-based container of data files that follow a specific Model

    Dataset is a generic class that cannot be instantiated directly. Instead, a Dataset class needs
    to be specialized with a data model before Dataset objects can be instantiated. A data model
    functions as a data parser and guarantees that the parsed data follows the specified model.

    The specialization must be done through the use of Model, either directly, e.g.::

        MyDataset = Dataset[Model[dict[str, list[int]]])

    ... or indirectly, using a Model subclass, e.g.::

        class MyModel(Model[dict[str, list[int]]):
            pass

        MyDataset = Dataset[MyModel]

    ... alternatively through the specification of a Dataset subclass::

        class MyDataset(Dataset[MyModel]):
            pass

    The specialization can also be done in a more deeply nested structure, e.g.::

        class MyNumberList(Model[list[int]]):
            pass

        class MyToplevelDict(Model[dict[str, MyNumberList]]):
            pass

        class MyDataset(Dataset[MyToplevelDict]):
            pass

    Once instantiated, a dataset object functions as a dict of data files, with the keys
    referring to the data file names and the contents to the data file contents, e.g.::

        MyNumberListDataset = Dataset[Model[list[int]]]

        my_dataset = MyNumberListDataset({'file_1': [1,2,3]})
        my_dataset['file_2'] = [2,3,4]

        print(my_dataset.keys())

    The Dataset class is a wrapper class around the powerful `GenericModel` class from pydantic.
    """
    class Config:
        validate_assignment = True
        # json_loads = orjson.loads
        # json_dumps = orjson_dumps

    data: dict[str, ModelT] = Field(default={})

    def __class_getitem__(cls, model: ModelT) -> ModelT:
        # TODO: change model type to params: Type[Any] | tuple[Type[Any], ...]
        #       as in GenericModel.

        # For now, only singular model types are allowed. These lines are needed for
        # interoperability with pydantic GenericModel, which internally stores the model
        # as a tuple:
        if isinstance(model, tuple) and len(model) == 1:
            model = model[0]

        orig_model = model

        model = cls._origmodel_if_annotated_optional(model)

        if not isinstance(model, TypeVar) \
                and not lenient_issubclass(model, Model) \
                and not is_strict_subclass(cls, Dataset):
            raise TypeError('Invalid model: {}! '.format(model)
                            + 'omnipy Dataset models must be a specialization of the omnipy '
                            'Model class.')

        if cls == Dataset and not is_optional(model):
            model = Annotated[Optional[model], 'Fake Optional from Dataset']

        created_dataset = super().__class_getitem__(model)

        _cleanup_name_qualname_and_module(cls, created_dataset, model, orig_model)

        return created_dataset

    def __init__(
        self,
        value: dict[str, object] | Iterator[tuple[str, object]] | UndefinedType = Undefined,
        *,
        data: dict[str, object] | UndefinedType = Undefined,
        **input_data: object,
    ) -> None:
        # TODO: Error message when forgetting parenthesis when creating Dataset should be improved.
        #       Unclear where this can be done, if anywhere? E.g.:
        #           a = Dataset[Model[int]]
        #           a['adsfas'] = 2
        #           Traceback (most recent call last):
        #             ...
        #           TypeError: 'ModelMetaclass' object does not support item assignment
        #
        # TODO: Disallow e.g.:
        #       Dataset[Model[str]](Model[int](5)) ==  Dataset[Model[str]](data=Model[int](5))
        #       == Dataset[Model[str]](data={'__root__': Model[str]('5')})

        super_kwargs = {}

        assert DATA_KEY not in input_data, \
            ('Not allowed with"data" as input_data key. Not sure how you managed this? Are you '
             'trying to break Dataset init on purpose?')

        if value != Undefined:
            assert data == Undefined, \
                'Not allowed to combine positional and "data" keyword argument'
            assert len(input_data) == 0, 'Not allowed to combine positional and keyword arguments'
            super_kwargs[DATA_KEY] = value

        if data != Undefined:
            assert len(input_data) == 0, \
                "Not allowed to combine 'data' with other keyword arguments"
            super_kwargs[DATA_KEY] = data

        if DATA_KEY not in super_kwargs and len(input_data) > 0:
            super_kwargs[DATA_KEY] = input_data

        if self.get_model_class() == ModelT:
            self._raise_no_model_exception()

        GenericModel.__init__(self, **super_kwargs)
        UserDict.__init__(self, self.data)  # noqa
        if not self.__doc__:
            self._set_standard_field_description()

    @classmethod
    def get_model_class(cls) -> Type[Model]:
        """
        Returns the concrete Model class used for all data files in the dataset, e.g.:
        `Model[list[int]]`
        :return: The concrete Model class used for all data files in the dataset
        """
        model_type = cls.__fields__.get(DATA_KEY).type_
        return cls._origmodel_if_annotated_optional(model_type)

    @classmethod
    def _origmodel_if_annotated_optional(cls, model):
        if get_origin(model) is Annotated:
            unannotated_type = get_args(model)[0]
            if is_optional(unannotated_type):
                model = get_args(unannotated_type)[0]
        return model

    # TODO: Update _raise_no_model_exception() text. Model is now a requirement
    @staticmethod
    def _raise_no_model_exception() -> None:
        raise TypeError(
            'Note: The Dataset class requires a concrete model to be specified as '
            'a type hierarchy within brackets either directly, e.g.:\n\n'
            '\tmodel = Dataset[list[int]]()\n\n'
            'or indirectly in a subclass definition, e.g.:\n\n'
            '\tclass MyNumberListDataset(Dataset[list[int]]): ...\n\n'
            'In both cases, the use of the Model class or a subclass is encouraged if anything '
            'other than the simplest cases, e.g.:\n\n'
            '\tclass MyNumberListModel(Model[list[int]]): ...\n'
            '\tclass MyDataset(Dataset[MyNumberListModel]): ...\n\n'
            'Usage of Dataset without a type specification results in this exception. '
            'Similar use of the Model class do not currently result in an exception, only '
            'a warning message the first time this is done. However, this is just a '
            '"poor man\'s exception" due to complex technicalities in that class. Please '
            'explicitly specify types in both cases. ')

    def _set_standard_field_description(self) -> None:
        self.__fields__[DATA_KEY].field_info.description = self._get_standard_field_description()

    @classmethod
    def _get_standard_field_description(cls) -> str:
        return ('This class represents a data in the `omnipy` Python package and contains '
                'a set of named data items that follows the same data model. '
                'It is a statically typed specialization of the Dataset class according to a '
                'particular specialization of the Model class. Both main classes are wrapping '
                'the excellent Python package named `pydantic`.')

    def __setitem__(self, data_file: str, data_obj: Any) -> None:
        has_prev_value = data_file in self.data
        prev_value = self.data.get(data_file)

        try:
            self.data[data_file] = data_obj
            self._validate(data_file)
        except:  # noqa
            if has_prev_value:
                self.data[data_file] = prev_value
            else:
                del self.data[data_file]
            raise

    def __getitem__(self, data_file: str) -> Any:
        if data_file in self.data:
            return self.data[data_file]
        else:
            return self.data[data_file]

    @classmethod
    def update_forward_refs(cls, **localns: Any) -> None:
        """
        Try to update ForwardRefs on fields based on this Model, globalns and localns.
        """
        cls.get_model_class().update_forward_refs(**localns)  # Update Model cls
        super().update_forward_refs(**localns)
        cls.__name__ = remove_forward_ref_notation(cls.__name__)
        cls.__qualname__ = remove_forward_ref_notation(cls.__qualname__)

    def _validate(self, _data_file: str) -> None:
        self.data = self.data  # Triggers pydantic validation, as validate_assignment=True

    def __iter__(self) -> Iterator:
        return UserDict.__iter__(self)

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in self.__dict__ or attr == DATA_KEY or attr.startswith('__'):
            super().__setattr__(attr, value)
        else:
            raise RuntimeError('Model does not allow setting of extra attributes')

    @root_validator()
    def _parse_root_object(cls, root_obj: dict[str, ModelT]) -> Any:  # noqa
        assert DATA_KEY in root_obj
        data_dict = root_obj[DATA_KEY]
        model = cls.get_model_class()
        for key, val in data_dict.items():
            if val is None:
                data_dict[key] = model.parse_obj(val)
        return {DATA_KEY: data_dict}

    def to_data(self) -> dict[str, Any]:
        return GenericModel.dict(self).get(DATA_KEY)

    def from_data(self,
                  data: dict[str, Any] | Iterator[tuple[str, Any]],
                  update: bool = True) -> None:
        if not isinstance(data, dict):
            data = dict(data)

        if not update:
            self.clear()

        for data_file, obj_val in data.items():
            new_model = self.get_model_class()()  # noqa
            new_model.from_data(obj_val)
            self[data_file] = new_model

    def absorb(self, other: 'Dataset'):
        self.from_data(other.to_data(), update=True)

    def absorb_and_replace(self, other: 'Dataset'):
        self.from_data(other.to_data(), update=False)

    def to_json(self, pretty=True) -> dict[str, str]:
        result = {}

        for key, val in self.to_data().items():
            result[key] = self._pretty_print_json(val) if pretty else json.dumps(val)
        return result

    def from_json(self,
                  data: dict[str, str] | Iterator[tuple[str, str]],
                  update: bool = True) -> None:
        if not isinstance(data, dict):
            data = dict(data)

        if not update:
            self.clear()

        for data_file, obj_val in data.items():
            new_model = self.get_model_class()()  # noqa
            new_model.from_json(obj_val)
            self[data_file] = new_model

    # @classmethod
    # def get_type_args(cls):
    #     return cls.__fields__.get(DATA_KEY).type_
    #
    #
    # @classmethod
    # def create_from_json(cls, data: str, tuple[str]]):
    #     if isinstance(data, tuple):
    #         data = data[0]
    #
    #     obj = cls()
    #     obj.from_json(data, update=False)
    #     return obj
    #
    # def __reduce__(self):
    #     return self.__class__.create_from_json, (self.to_json(),)

    @classmethod
    def to_json_schema(cls, pretty=True) -> str | dict[str, str]:
        result = {}
        schema = cls.schema()
        for key, val in schema['properties']['data'].items():
            result[key] = val
        result['title'] = schema['title']
        result['definitions'] = schema['definitions']
        if pretty:
            return cls._pretty_print_json(result)
        else:
            return json.dumps(result)

    @staticmethod
    def _pretty_print_json(json_content: Any) -> str:
        return json.dumps(json_content, indent=2)

    def save(self, directory: str):
        serializer_registry = self._get_serializer_registry()

        parsed_dataset, serializer = serializer_registry.auto_detect_tar_file_serializer(self)

        if serializer is None:
            print(f'Unable to find a serializer for dataset with data type "{type(self)}". '
                  f'Will abort saving...')
        else:
            print(f'Writing dataset as a gzipped tarpack to "{os.path.abspath(directory)}"')

            out_tar_gz_path = f'{directory}.tar.gz'

            with open(out_tar_gz_path, 'wb') as out_tar_gz_file:
                out_tar_gz_file.write(serializer.serialize(parsed_dataset))

            if not os.path.exists(directory):
                os.makedirs(directory)

            tar = tarfile.open(out_tar_gz_path)
            tar.extractall(path=directory)
            tar.close()

    def load(self, directory: str):
        serializer_registry = self._get_serializer_registry()
        return serializer_registry.load_from_tar_file_path(self, directory, self)

    @staticmethod
    def _get_serializer_registry():
        from omnipy import runtime
        if len(runtime.objects.serializers.serializers) == 0:
            from omnipy.modules import register_serializers
            register_serializers(runtime.objects.serializers)
        serializer_registry = runtime.objects.serializers
        return serializer_registry

    def as_multi_model_dataset(self) -> 'MultiModelDataset[ModelT]':
        multi_model_dataset = MultiModelDataset[self.get_model_class()]()
        for data_file in self:
            multi_model_dataset.data[data_file] = self.data[data_file]
        return multi_model_dataset

    def __eq__(self, other: object) -> bool:
        # return self.__class__ == other.__class__ and super().__eq__(other)
        return isinstance(other, Dataset) \
            and self.__class__ == other.__class__ \
            and self.data == other.data \
            and self.to_data() == other.to_data()  # last is probably unnecessary, but just in case

    def __repr__(self) -> str:
        return super().__repr__()

    def __repr_args__(self):
        return [(k, v.contents) for k, v in self.data.items()]


# TODO: Use json serializer package from the pydantic config instead of 'json'


class MultiModelDataset(Dataset[ModelT], Generic[ModelT]):
    """
        Variant of Dataset that allows custom models to be set on individual data files

        Note that the general model still needs to hold for all data files, in addition to any
        custom models.
    """

    _custom_field_models: dict[str, ModelT] = PrivateAttr(default={})

    def set_model(self, data_file: str, model: ModelT) -> None:
        try:
            self._custom_field_models[data_file] = model
            if data_file in self.data:
                self._validate(data_file)
            else:
                self.data[data_file] = model()
        except ValidationError:
            del self._custom_field_models[data_file]
            raise

    def get_model(self, data_file: str) -> ModelT:
        if data_file in self._custom_field_models:
            return self._custom_field_models[data_file]
        else:
            return self.get_model_class()

    def _validate(self, data_file: str) -> None:
        if data_file in self._custom_field_models:
            model = self._custom_field_models[data_file]
            if not isinstance(model, Model):
                model = Model[model]
            data_obj = self._to_data_if_model(self.data[data_file])
            parsed_data = self._to_data_if_model(model(data_obj))
            self.data[data_file] = parsed_data
        super()._validate(data_file)  # validates all data according to ModelT

    @staticmethod
    def _to_data_if_model(data_obj: Any):
        if isinstance(data_obj, Model):
            data_obj = data_obj.to_data()
        return data_obj



from typing import Generic, TypeVar

from omnipy.data.dataset import Dataset
from omnipy.modules.general.models import (NestedFrozenDictsModel,
                                           NestedFrozenDictsOrTuplesModel,
                                           NestedFrozenTuplesModel)

_KeyT = TypeVar('_KeyT')
_ScT = TypeVar('_ScT')


class NestedFrozenTuplesDataset(Dataset[NestedFrozenTuplesModel[_ScT]], Generic[_ScT]):
    ...


class NestedFrozenDictsDataset(Dataset[NestedFrozenDictsModel[_KeyT, _ScT]], Generic[_KeyT, _ScT]):
    ...


class NestedFrozenDictsOrTuplesDataset(Dataset[NestedFrozenDictsOrTuplesModel[_KeyT, _ScT]],
                                       Generic[_KeyT, _ScT]):
    ...



__version__ = '0.12.3'

import os
import sys
from typing import Optional

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model
from omnipy.hub.runtime import Runtime

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# TODO: The check disabling runtime for tests also trigger for tests that are run outside of Omnipy,
#  breaking tests on the user side.
#  Find a better way to disable the global runtime object for Omnipy tests


def _get_runtime() -> Optional['Runtime']:
    if 'pytest' not in sys.modules:
        return Runtime()
    else:
        return None


runtime: Optional['Runtime'] = _get_runtime()

__all__ = [Model, Dataset]



from copy import deepcopy

from omnipy.compute.task import TaskTemplate
from omnipy.compute.typing import mypy_fix_task_template
from omnipy.modules.json.datasets import JsonListOfDictsDataset


@mypy_fix_task_template
@TaskTemplate()
def remove_columns(json_dataset: JsonListOfDictsDataset,
                   column_keys_for_data_items: dict[str, list[str]]) -> JsonListOfDictsDataset:
    # TODO: implement general solution to make sure that one does not modify input data by
    #       automatically copying or otherwise. Perhaps setting immutable/frozen option in pydantic,
    #       if available?
    #
    dataset_copy = deepcopy(json_dataset)
    for data_item_key, column_keys in column_keys_for_data_items.items():
        for record in dataset_copy[data_item_key]:
            for column in column_keys:
                if column in record:
                    del record[column]
    return JsonListOfDictsDataset(json_dataset.to_data())


# @mypy_fix_task_template
# @TaskTemplate()
# def convert_to_1nf(input_dataset: Dataset[TableOfStringsAndLists]) -> Dataset[TableOfStrings]:
#     out_dataset = Dataset[TableOfStrings]()
#     for table_name, table in input_dataset.items():
#         # for record in table:
#         #     ...
#         out_dataset[table_name] = table
#     return out_dataset



from omnipy.api.protocols.public.data import IsSerializerRegistry


def register_serializers(registry: IsSerializerRegistry):
    from .json.serializers import JsonDatasetToTarFileSerializer
    from .pandas.serializers import PandasDatasetToTarFileSerializer
    from .raw.serializers import RawDatasetToTarFileSerializer

    registry.register(PandasDatasetToTarFileSerializer)
    registry.register(RawDatasetToTarFileSerializer)
    registry.register(JsonDatasetToTarFileSerializer)


# TODO: Add module with helper classes/functions/takss to make it simpler to contact REST apis
#       Augmentation service should have some attempts at this.



from dataclasses import dataclass, field
import locale as pkg_locale
import logging
from pathlib import Path

from omnipy.api.typedefs import LocaleType


def _get_log_dir_path() -> str:
    return str(Path.cwd().joinpath(Path('logs')))


@dataclass
class RootLogConfig:
    log_format_str: str = '{engine} {asctime} - {levelname}: {message} [{name}]'
    locale: LocaleType = pkg_locale.getlocale()
    log_to_stdout: bool = True
    log_to_stderr: bool = True
    log_to_file: bool = True
    stdout_log_min_level: int = logging.INFO
    stderr_log_min_level: int = logging.ERROR
    file_log_min_level: int = logging.WARNING
    file_log_dir_path: str = field(default_factory=_get_log_dir_path)



from typing import NamedTuple


class MethodInfo(NamedTuple):
    state_changing: bool
    maybe_returns_same_type: bool


# Ordered after their order in the Python documentation
# (https://docs.python.org/3.10/reference/datamodel.html)
SPECIAL_METHODS_INFO: dict[str, MethodInfo] = {
    # 3.3.1. Basic customization ############################################
    '__bool__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    # 3.3.7. Emulating container types ######################################
    '__len__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__length_hint__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__getitem__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__setitem__': MethodInfo(state_changing=True, maybe_returns_same_type=False),
    '__delitem__': MethodInfo(state_changing=True, maybe_returns_same_type=False),
    '__missing__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__iter__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__reversed__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__contains__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    # 3.3.8. Emulating numeric types ########################################
    # - Binary arithmetic ---------------------------------------------------
    '__add__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__sub__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__mul__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__matmul__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__truediv__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__floordiv__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__mod__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__divmod__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__pow__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__lshift__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rshift__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__and__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__xor__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__or__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    # - Binary arithmetic (right-hand side) ---------------------------------
    '__radd__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rsub__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rmul__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rmatmul__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rtruediv__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rfloordiv__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rmod__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rdivmod__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rpow__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rlshift__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rrshift__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rand__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__rxor__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__ror__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    # - Binary arithmetic (in-place, e.g. '=+') -----------------------------
    '__iadd__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__isub__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__imul__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__imatmul__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__itruediv__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__ifloordiv__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__imod__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__ipow__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__ilshift__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__irshift__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__iand__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__ixor__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    '__ior__': MethodInfo(state_changing=True, maybe_returns_same_type=True),
    # - Unary arithmetic operations (e.g. -x) -------------------------------
    '__neg__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__pos__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__abs__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__invert__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    # - Converting to the appropriate type ----------------------------------
    '__complex__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__int__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    '__float__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    # - Used to implement operator.index() - always integer -----------------
    '__index__': MethodInfo(state_changing=False, maybe_returns_same_type=False),
    # - Rounding operations - mostly integers -------------------------------
    '__round__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__trunc__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__floor__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
    '__ceil__': MethodInfo(state_changing=False, maybe_returns_same_type=True),
}



import os

from omnipy.data.model import Model


class SplitLinesModel(Model[str | list[str]]):
    @classmethod
    def _parse_data(cls, data: str | list[str]) -> list[str]:
        if isinstance(data, list):
            return data
        return data.split(os.linesep)


class SplitAndStripLinesModel(SplitLinesModel):
    @classmethod
    def _parse_data(cls, data: str | list[str]) -> list[str]:
        return [line.strip() for line in super()._parse_data(data)]


class JoinLinesModel(Model[list[str] | str]):
    @classmethod
    def _parse_data(cls, data: list[str] | str) -> str:
        if isinstance(data, str):
            return data
        return os.linesep.join(data)



from dataclasses import dataclass, field

from omnipy.api.protocols.public.hub import IsRuntime
from omnipy.util.publisher import DataPublisher


@dataclass
class RuntimeEntryPublisher(DataPublisher):
    _back: IsRuntime | None = field(default=None, init=False, repr=False)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

        if hasattr(self, key) and not key.startswith('_') and self._back is not None:
            self._back.reset_subscriptions()



from dataclasses import dataclass, field
import logging
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler
import os
from pathlib import Path
from sys import stderr, stdout

from omnipy.api.protocols.public.config import IsRootLogConfig
from omnipy.config.root_log import RootLogConfig
from omnipy.hub.entry import RuntimeEntryPublisher
from omnipy.util.helpers import get_datetime_format


@dataclass
class RootLogConfigEntryPublisher(RootLogConfig, RuntimeEntryPublisher):
    ...


@dataclass
class RootLogObjects:
    _config: IsRootLogConfig = field(
        init=False, repr=False, default_factory=RootLogConfigEntryPublisher)

    formatter: logging.Formatter | None = None
    stdout_handler: StreamHandler | None = None
    stderr_handler: StreamHandler | None = None
    file_handler: TimedRotatingFileHandler | None = None

    def __post_init__(self):
        self._configure_all_objects()

    def set_config(self, config: IsRootLogConfig):
        self._config = config
        self._configure_all_objects()

    def _configure_all_objects(self):
        self._remove_all_handlers_from_root_logger()

        self._configure_formatter()
        self._configure_stdout_handler()
        self._configure_stderr_handler()
        self._configure_file_handler()

        self._add_all_handlers_to_root_logger()

    def _configure_formatter(self):
        if self._config.log_format_str:
            datetime_fmt = get_datetime_format(self._config.locale)
            self.formatter = logging.Formatter(self._config.log_format_str, datetime_fmt, style='{')
        else:
            self.formatter = None

    def _configure_stdout_handler(self):
        if self._config.log_to_stdout:
            config = self._config

            class StdErrBasedMaxLevelFilter(logging.Filter):
                def filter(self, record):
                    return record.levelno < config.stderr_log_min_level

            self.stdout_handler = StreamHandler(stdout)
            self.stdout_handler.setLevel(self._config.stdout_log_min_level)
            if self._config.log_to_stderr:
                self.stdout_handler.addFilter(StdErrBasedMaxLevelFilter())
        else:
            self.stdout_handler = None

    def _configure_stderr_handler(self) -> StreamHandler | None:
        if self._config.log_to_stderr:
            self.stderr_handler = StreamHandler(stderr)
            self.stderr_handler.setLevel(self._config.stderr_log_min_level)
        else:
            self.stderr_handler = None

    def _configure_file_handler(self) -> TimedRotatingFileHandler | None:
        if self._config.log_to_file:
            log_dir_path = self._config.file_log_dir_path
            if not os.path.exists(log_dir_path):
                os.makedirs(log_dir_path)

            log_file_path = Path(log_dir_path).joinpath('omnipy.log')
            self.file_handler = TimedRotatingFileHandler(
                log_file_path, when='d', interval=1, backupCount=7)
            self.file_handler.setLevel(self._config.file_log_min_level)
        else:
            self.file_handler = None

    def _remove_all_handlers_from_root_logger(self):
        root_logger = logging.root
        handlersToRemove = [
            handler for handler in root_logger.handlers
            if isinstance(handler, StreamHandler) or isinstance(handler, TimedRotatingFileHandler)
        ]
        for handler in handlersToRemove:
            root_logger.removeHandler(handler)

    def _add_handler_to_root_logger(self, handler: logging.Handler | None):
        if handler:
            root_logger = logging.root
            if self.formatter:
                handler.setFormatter(self.formatter)
            root_logger.addHandler(handler)

    def _add_all_handlers_to_root_logger(self):
        self._add_handler_to_root_logger(self.stdout_handler)
        self._add_handler_to_root_logger(self.stderr_handler)
        self._add_handler_to_root_logger(self.file_handler)



from dataclasses import dataclass, field
import os
from pathlib import Path

from omnipy.api.enums import (ConfigOutputStorageProtocolOptions,
                              ConfigPersistOutputsOptions,
                              ConfigRestoreOutputsOptions)
from omnipy.api.protocols.public.config import (IsLocalOutputStorage,
                                                IsOutputStorage,
                                                IsS3OutputStorage)


def _get_persist_data_dir_path() -> str:
    return str(Path.cwd().joinpath(Path('outputs')))


@dataclass
class LocalOutputStorage:
    persist_data_dir_path: str = field(default_factory=_get_persist_data_dir_path)


@dataclass
class S3OutputStorage:
    persist_data_dir_path: str = os.path.join('omnipy', 'outputs')
    endpoint_url: str = ''
    bucket_name: str = ''
    access_key: str = ''
    secret_key: str = ''


@dataclass
class OutputStorage:
    persist_outputs: ConfigPersistOutputsOptions = \
        ConfigPersistOutputsOptions.ENABLE_FLOW_AND_TASK_OUTPUTS
    restore_outputs: ConfigRestoreOutputsOptions = \
        ConfigRestoreOutputsOptions.DISABLED
    protocol: ConfigOutputStorageProtocolOptions = ConfigOutputStorageProtocolOptions.LOCAL
    local: IsLocalOutputStorage = field(default_factory=LocalOutputStorage)
    s3: IsS3OutputStorage = field(default_factory=S3OutputStorage)


@dataclass
class JobConfig:
    output_storage: IsOutputStorage = field(default_factory=OutputStorage)



from collections.abc import Hashable, Iterable
from copy import copy, deepcopy
import inspect
import locale as pkg_locale
from types import GenericAlias, UnionType
from typing import (Annotated,
                    Any,
                    cast,
                    ClassVar,
                    get_args,
                    get_origin,
                    Mapping,
                    NamedTuple,
                    Protocol,
                    Type,
                    TypeVar,
                    Union)

from isort import place_module
from isort.sections import STDLIB
from pydantic import ValidationError
from pydantic.typing import display_as_type
from typing_inspect import get_generic_bases, is_generic_type

from omnipy.api.typedefs import LocaleType

KeyT = TypeVar('KeyT', bound=Hashable)

Dictable = Mapping[KeyT, Any] | Iterable[tuple[KeyT, Any]]


def as_dictable(obj: object) -> Dictable | None:
    def _is_iterable_of_tuple_pairs(obj_inner: object) -> bool:
        return isinstance(obj_inner, Iterable) and \
            all(isinstance(el, tuple) and len(el) == 2 for el in obj_inner)

    if isinstance(obj, Mapping) or _is_iterable_of_tuple_pairs(obj):
        return cast(Dictable, obj)
    else:
        return None


def create_merged_dict(dictable_1: Dictable[KeyT], dictable_2: Dictable[KeyT]) -> dict[KeyT, Any]:
    merged_dict = dictable_1 if isinstance(dictable_1, dict) else dict(dictable_1)
    dict_2 = dictable_2 if isinstance(dictable_2, dict) else dict(dictable_2)
    merged_dict |= dict_2
    return merged_dict


def remove_none_vals(**kwargs: object) -> dict[object, object]:
    return {key: val for key, val in kwargs.items() if val is not None}


def get_datetime_format(locale: LocaleType | None = None) -> str:
    pkg_locale.setlocale(pkg_locale.LC_ALL, locale)

    if hasattr(pkg_locale, 'nl_langinfo'):  # noqa
        datetime_format = pkg_locale.nl_langinfo(pkg_locale.D_T_FMT)
    else:
        datetime_format = '%a %b %e %X %Y'
    return datetime_format


async def resolve(val):
    return await val if inspect.isawaitable(val) else val


def repr_max_len(data: object, max_len: int = 200):
    repr_str = repr(data)
    return f'{repr_str[:max_len]}...' if len(repr_str) > max_len else repr_str


def get_bases(cls):
    return get_generic_bases(cls) if is_generic_type(cls) else cls.__bases__


def generic_aware_issubclass_ignore_args(cls, cls_or_tuple):
    try:
        return issubclass(cls, cls_or_tuple)
    except TypeError:
        return issubclass(get_origin(cls), cls_or_tuple)


def transfer_generic_args_to_cls(to_cls, from_generic_type):
    try:
        return to_cls[get_args(from_generic_type)]
    except (TypeError, AttributeError):
        return to_cls


def is_iterable(obj: object) -> bool:
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def is_union(cls_or_type: type | UnionType | None | object) -> bool:
    return get_origin(cls_or_type) in [Union, UnionType]


def is_optional(cls_or_type: type | UnionType | None | object) -> bool:
    return is_union(cls_or_type) and type(None) in get_args(cls_or_type)


def all_equals(first, second) -> bool:
    equals = first == second
    if is_iterable(equals):
        if hasattr(equals, 'all') and callable(getattr(equals, 'all')):
            return equals.all(None)  # works for both pandas and numpy
        else:
            return all(equals)
    else:
        return equals


def is_strict_subclass(
        __cls: type,
        __class_or_tuple: type | UnionType | tuple[type | UnionType | tuple[Any, ...], ...]
) -> bool:
    if issubclass(__cls, __class_or_tuple):
        if isinstance(__class_or_tuple, Iterable):
            return __cls not in __class_or_tuple
        else:
            return __cls != __class_or_tuple
    return False


class IsDataclass(Protocol):
    __dataclass_fields__: ClassVar[dict]


def ensure_plain_type(in_type: type | GenericAlias) -> type | GenericAlias | None | Any:
    return get_origin(in_type) if get_args(in_type) else in_type


def remove_annotated_plus_optional_if_present(
        type_or_class: Type | UnionType | object) -> Type | UnionType | object:
    if get_origin(type_or_class) == Annotated:
        type_or_class = get_args(type_or_class)[0]
        if is_optional(type_or_class):
            args = get_args(type_or_class)
            if len(args) == 2:
                type_or_class = args[0]
            else:
                type_or_class = Union[args[:-1]]
    return type_or_class


def remove_forward_ref_notation(type_str: str):
    return type_str.replace("ForwardRef('", '').replace("')", '')


def generate_qualname(cls_name: str, model: Any) -> str:
    m_module = model.__module__ if hasattr(model, '__module__') else ''
    m_module_prefix = f'{m_module}.' if m_module and place_module(m_module) != STDLIB else ''
    fully_qual_model_name = f'{m_module_prefix}{display_as_type(model)}'
    return f'{cls_name}[{fully_qual_model_name}]'


class Snapshot(NamedTuple):
    id: int
    obj_copy: object


class RestorableContents:
    def __init__(self):
        self._last_snapshot: Snapshot | None = None

    def has_snapshot(self) -> bool:
        return self._last_snapshot is not None

    def take_snapshot(self, obj: object):
        try:
            snapshot_obj = deepcopy(obj)
        except (TypeError, ValueError, ValidationError):
            snapshot_obj = copy(obj)
        self._last_snapshot = Snapshot(id(obj), snapshot_obj)

    def _assert_not_empty(self):
        assert self.has_snapshot(), 'No snapshot has been taken yet'

    def get_last_snapshot(self) -> object:
        self._assert_not_empty()
        return self._last_snapshot.obj_copy

    def last_snapshot_taken_of_same_obj(self, obj: object) -> bool:
        self._assert_not_empty()
        return self._last_snapshot.id == id(obj)

    def differs_from_last_snapshot(self, obj: object) -> bool:
        self._assert_not_empty()
        return not all_equals(self._last_snapshot.obj_copy, obj)



from omnipy.data.model import Model
from omnipy.modules.json.models import JsonDictM, JsonListM


class TableOfStrings(Model[list[dict[str, str]]]):
    ...


class JsonTableOfStrings(JsonListM[JsonDictM[str]]):
    ...


class TableOfStringsAndLists(Model[list[dict[str, str | list[str]]]]):
    ...



from contextlib import AbstractContextManager
from typing import Callable, ParamSpec, TypeVar

DecoratedP = ParamSpec('DecoratedP')
DecoratedR = TypeVar('DecoratedR')
CallbackP = ParamSpec('CallbackP')


def add_callback_after_call(func: Callable[DecoratedP, DecoratedR],
                            callback_func: Callable[CallbackP, None],
                            *cb_args: CallbackP.args,
                            with_context: AbstractContextManager | None = None,
                            **cb_kwargs: CallbackP.kwargs) -> Callable[DecoratedP, DecoratedR]:
    class ValidateAfterCall(AbstractContextManager):
        def __enter__(self):
            ...

        def __exit__(self, exc_type, exc_val, exc_tb):
            ret = None

            if exc_val is None:
                callback_func(*cb_args, **cb_kwargs)
            return ret

    def _inner(*args: DecoratedP.args, **kwargs: DecoratedP.kwargs) -> DecoratedR:
        if with_context:
            with with_context:
                with ValidateAfterCall():
                    return func(*args, **kwargs)
        with ValidateAfterCall():
            return func(*args, **kwargs)

    return _inner



from collections.abc import Iterable
from typing import Any

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model, ROOT_KEY

from . import pd


class PandasModel(Model[pd.DataFrame]):
    @classmethod
    def _parse_data(cls, data: pd.DataFrame) -> pd.DataFrame:
        cls._data_column_names_are_strings(data)
        cls._data_not_empty_object(data)
        return data

    @staticmethod
    def _data_column_names_are_strings(data: pd.DataFrame) -> None:
        for column in data.columns:
            assert isinstance(column, str)

    @staticmethod
    def _data_not_empty_object(data: pd.DataFrame) -> None:
        assert not any(data.isna().all(axis=1))

    def dict(self, *args, **kwargs) -> dict[str, dict[Any, Any]]:
        df = super().dict(*args, **kwargs)[ROOT_KEY]
        df = df.replace({pd.NA: None})
        return {ROOT_KEY: df.to_dict(orient='records')}

    def from_data(self, value: Iterable[Any]) -> None:
        self._validate_and_set_contents(pd.DataFrame(value).convert_dtypes())

    def from_json(self, value: str) -> None:
        self._validate_and_set_contents(pd.read_json(value).convert_dtypes())


class PandasDataset(Dataset[PandasModel]):
    ...


class ListOfPandasDatasetsWithSameNumberOfFiles(Model[list[PandasDataset]]):
    @classmethod
    def _parse_data(cls, dataset_list: list[PandasDataset]) -> Any:
        assert len(dataset_list) >= 2
        assert all(len(dataset) for dataset in dataset_list)



from datetime import datetime
import logging
from logging import getLogger, INFO, Logger
import time


class LogMixin:
    def __init__(self) -> None:
        self._logger: Logger = getLogger(f'{self.__class__.__module__}.{self.__class__.__name__}')
        self._logger.setLevel(INFO)

    @property
    def logger(self) -> Logger:
        return self._logger

    def log(self, log_msg: str, level: int = INFO, datetime_obj: datetime | None = None):
        if self._logger is not None:
            create_time = time.mktime(datetime_obj.timetuple()) if datetime_obj else time.time()
            _former_log_record_factory = logging.getLogRecordFactory()
            if _former_log_record_factory.__name__ != '_log_record_editor':

                def _log_record_editor(*args, **kwargs):
                    record = _former_log_record_factory(*args, **kwargs)
                    record.created = create_time
                    record.engine = f"[{record.name.split('.')[0].upper()}]"
                    if len(record.engine) < 9:
                        record.engine += ' '
                    return record

                logging.setLogRecordFactory(_log_record_editor)

            self._logger.log(level, log_msg)



from collections.abc import Mapping, Sequence
import functools
import inspect
import json
from types import UnionType
from typing import (Annotated,
                    Any,
                    cast,
                    Generic,
                    get_args,
                    get_origin,
                    Optional,
                    SupportsIndex,
                    Type,
                    TypeVar,
                    Union)

# from orjson import orjson
from pydantic import NoneIsNotAllowedError
from pydantic import Protocol as PydanticProtocol
from pydantic import root_validator, ValidationError
from pydantic.fields import ModelField, Undefined, UndefinedType
from pydantic.generics import GenericModel
from pydantic.main import ModelMetaclass, validate_model
from pydantic.typing import display_as_type, is_none_type
from pydantic.utils import lenient_isinstance, lenient_issubclass

from omnipy.data.methodinfo import MethodInfo, SPECIAL_METHODS_INFO
from omnipy.util.contexts import AttribHolder, LastErrorHolder, nothing
from omnipy.util.decorators import add_callback_after_call
from omnipy.util.helpers import (all_equals,
                                 ensure_plain_type,
                                 generate_qualname,
                                 is_optional,
                                 is_union,
                                 remove_annotated_plus_optional_if_present,
                                 remove_forward_ref_notation,
                                 RestorableContents)

_KeyT = TypeVar('_KeyT')
_ValT = TypeVar('_ValT')
_IdxT = TypeVar('_IdxT', bound=SupportsIndex)
RootT = TypeVar('RootT', covariant=True, bound=object)

ROOT_KEY = '__root__'


def _cleanup_name_qualname_and_module(cls, created_model_or_dataset, model, orig_model):
    if isinstance(model, str):  # ForwardRef
        created_model_or_dataset.__name__ = f'{cls.__name__}[{model}]'
        created_model_or_dataset.__qualname__ = created_model_or_dataset.__name__
    else:
        if created_model_or_dataset.__name__.startswith(f'{cls.__name__}[') \
                and get_origin(model) is Annotated:
            created_model_or_dataset.__name__ = f'{cls.__name__}[{display_as_type(orig_model)}]'
            created_model_or_dataset.__qualname__ = generate_qualname(cls.__qualname__, orig_model)
        else:
            created_model_or_dataset.__qualname__ = generate_qualname(cls.__qualname__, model)

    created_model_or_dataset.__module__ = cls.__module__


# def orjson_dumps(v, *, default):
#     # orjson.dumps returns bytes, to match standard json.dumps we need to decode
#     return orjson.dumps(v, default=default).decode()

_restorable_content_cache: dict[int, RestorableContents] = {}


class MyModelMetaclass(ModelMetaclass):
    # Hack to overcome bug in pydantic/fields.py (v1.10.13), lines 636-641:
    #
    # if origin is None or origin is CollectionsHashable:
    #     # field is not "typing" object eg. Union, dict, list etc.
    #     # allow None for virtual superclasses of NoneType, e.g. Hashable
    #     if isinstance(self.type_, type) and isinstance(None, self.type_):
    #         self.allow_none = True
    #     return
    #
    # This hinders models (including pure pydantic BaseModels) to be properly considered as
    # subfields, e.g. in `list[MyModel]` as `get_origin(MyModel) is None`. Here, we want allow_none
    # to be set to True so that Model is allowed to validate a None value.
    #
    # TODO: Revisit the need for MyModelMetaclass hack in pydantic v2

    def __instancecheck__(self, instance: Any) -> bool:
        if instance is None:
            return True
        return super().__instancecheck__(instance)


class Model(GenericModel, Generic[RootT], metaclass=MyModelMetaclass):
    """
    A data model containing a value parsed according to the model.

    If no value is provided, the value is set to the default value of the data model, found by
    calling the model class without parameters, e.g. `int()`.

    Model is a generic class that cannot be instantiated directly. Instead, a Model class needs
    to be specialized with a data type before Model objects can be instantiated. A data model
    functions as a data parser and guarantees that the parsed data follows the specified model.

    Example data model specialized as a class alias::

        MyNumberList = Model[list[int]]

    ... alternatively as a Model subclass::

        class MyNumberList(Model[list[int]]):
            pass

    Once instantiated, a Model object functions as a parser, e.g.::

        my_number_list = MyNumberList([2,3,4])

        my_number_list.contents = ['3', 4, True]
        assert my_number_list.contents == [3,4,1]

    While the following should raise a `ValidationError`::

        my_number_list.contents = ['abc', 'def']

    The Model class is a wrapper class around the powerful `GenericModel` class from pydantic.

    See also docs of the Dataset class for more usage examples.
    """

    __root__: RootT | None

    class Config:
        arbitrary_types_allowed = True
        validate_all = True
        # validate_assignment = True
        smart_union = True
        # json_loads = orjson.loads
        # json_dumps = orjson_dumps

    @classmethod
    def _get_default_value_from_model(cls, model: Type[RootT]) -> RootT:  # noqa: C901
        if isinstance(model, TypeVar):
            if model.__bound__ is None:  # noqa
                raise TypeError('The TypeVar "{}" needs to be bound to a '.format(model.__name__)
                                + 'type that provides a default value when called')
            else:
                model = model.__bound__  # noqa
        origin_type = get_origin(model)
        args = get_args(model)

        if origin_type is Annotated:
            model = args[0]
            origin_type = get_origin(model)
            args = get_args(model)

        if origin_type in (None, ()):
            origin_type = model

        if origin_type in [Union, UnionType]:
            if any(is_none_type(arg) for arg in args):
                return None

            last_error_holder = LastErrorHolder()
            for arg in args:
                if callable(arg):
                    with last_error_holder:
                        return cls._get_default_value_from_model(arg)
            last_error_holder.raise_derived(TypeError(f'Cannot instantiate model "{model}".'))

        if origin_type is tuple:
            if args and Ellipsis not in args:
                return tuple(cls._get_default_value_from_model(arg) for arg in args)

        return origin_type()

    @classmethod
    def _populate_root_field(cls, model: Type[RootT]) -> None:
        default_val = cls._get_default_value_from_model(model)

        def get_default_val() -> RootT:
            return default_val

        if ROOT_KEY in cls.__config__.fields:
            cls.__config__.fields[ROOT_KEY]['default_factory'] = get_default_val
        else:
            cls.__config__.fields[ROOT_KEY] = {'default_factory': get_default_val}

        if not get_origin(model) == Annotated and not is_optional(model):
            model = Annotated[Optional[model], 'Fake Optional from Model']

        data_field = ModelField.infer(
            name=ROOT_KEY,
            value=Undefined,
            annotation=model,
            class_validators=None,
            config=cls.__config__)

        cls.__fields__[ROOT_KEY] = data_field
        cls.__annotations__[ROOT_KEY] = model

        return model

    @classmethod
    def _depopulate_root_field(cls):
        del cls.__config__.fields[ROOT_KEY]
        del cls.__fields__[ROOT_KEY]
        del cls.__annotations__[ROOT_KEY]

    def __class_getitem__(cls, model: Type[RootT] | TypeVar) -> Type[RootT] | TypeVar:
        # TODO: change model type to params: Type[Any], tuple[Type[Any], ...]]
        #       as in GenericModel

        # For now, only singular model types are allowed. These lines are needed for
        # interoperability with pydantic GenericModel, which internally stores the model
        # as a tuple:
        if isinstance(model, tuple) and len(model) == 1:
            model = model[0]

        orig_model = model

        # Populating the root field at runtime instead of providing a __root__ Field explicitly
        # is needed due to the inability of typing/pydantic to provide a dynamic default based on
        # the actual type. The following issue in mypy seems relevant:
        # https://github.com/python/mypy/issues/3737 (as well as linked issues)
        if cls == Model:  # Only for the base Model class
            model = cls._populate_root_field(model)

        created_model = super().__class_getitem__(model)

        # As long as models are not created concurrently, setting the class members temporarily
        # should not have averse effects
        # TODO: Check if we can move to explicit definition of __root__ field at the object
        #       level in pydantic 2.0 (when it is released)
        if cls == Model:
            cls._depopulate_root_field()

        _cleanup_name_qualname_and_module(cls, created_model, model, orig_model)

        outer_type = created_model._get_root_type(outer=True, with_args=True)
        outer_type_plain = created_model._get_root_type(outer=True, with_args=False)

        if inspect.isclass(outer_type_plain) or is_union(outer_type):
            for name, method_info in SPECIAL_METHODS_INFO.items():
                outer_types = get_args(outer_type) if is_union(outer_type) else [outer_type_plain]
                for type_to_support in outer_types:
                    if hasattr(type_to_support, name):
                        setattr(created_model,
                                name,
                                functools.partialmethod(cls._special_method, name, method_info))
                        break

        return created_model

    def __new__(cls, value: Any | UndefinedType = Undefined, **kwargs):
        model_not_specified = ROOT_KEY not in cls.__fields__
        if model_not_specified:
            cls._raise_no_model_exception()
        return super().__new__(cls)

    # TODO: Allow e.g. Model[str](Model[int](5)) == Model[str](Model[int](5).contents).
    #       Should then work the same as dataset
    def __init__(
        self,
        value: Any | UndefinedType = Undefined,
        *,
        __root__: Any | UndefinedType = Undefined,
        **data: Any,
    ) -> None:
        super_data: dict[str, RootT] = {}
        num_root_vals = 0

        if value is not Undefined:
            super_data[ROOT_KEY] = cast(RootT, value)
            num_root_vals += 1

        if __root__ is not Undefined:
            super_data[ROOT_KEY] = cast(RootT, __root__)
            num_root_vals += 1

        if data:
            super_data[ROOT_KEY] = cast(RootT, data)
            num_root_vals += 1

        assert num_root_vals <= 1, 'Not allowed to provide root data in more than one argument'

        if ROOT_KEY in super_data \
                and isinstance(super_data[ROOT_KEY], Model) \
                and not is_none_type(super_data[ROOT_KEY]):  # Consequence of MyModelMetaclass hack
            super_data[ROOT_KEY] = super_data[ROOT_KEY].to_data()

        super().__init__(**super_data)

        self._take_snapshot_of_validated_contents()  # initial snapshot

        if not self.__class__.__doc__:
            self._set_standard_field_description()

    def __del__(self):
        if id(self) in _restorable_content_cache:
            del _restorable_content_cache[id(self)]

    @staticmethod
    def _raise_no_model_exception() -> None:
        raise TypeError('Note: The Model class requires a concrete model to be specified as '
                        'a type hierarchy within brackets either directly, e.g.:\n\n'
                        '\tmodel = Model[list[int]]([1,2,3])\n\n'
                        'or indirectly in a subclass definition, e.g.:\n\n'
                        '\tclass MyNumberList(Model[list[int]]): ...\n\n')

    def _set_standard_field_description(self) -> None:
        self.__fields__[ROOT_KEY].field_info.description = self._get_standard_field_description()

    @classmethod
    def _get_standard_field_description(cls) -> str:
        return ('This class represents a concrete model for data items in the `omnipy` Python '
                'package. It is a statically typed specialization of the Model class, '
                'which is itself wrapping the excellent Python package named `pydantic`.')

    @classmethod
    def validate(cls: Type['Model'], value: Any) -> 'Model':
        """
        Hack to allow overwriting of __iter__ method without compromising pydantic validation. Part
        of the pydantic API and not the Omnipy API.
        """
        if isinstance(value, Model):
            with AttribHolder(
                    value, '__iter__', GenericModel.__iter__, switch_to_other=True, on_class=True):
                return super().validate(value)
        else:
            return super().validate(value)

    @classmethod
    def update_forward_refs(cls, **localns: Any) -> None:
        """
        Try to update ForwardRefs on fields based on this Model, globalns and localns.
        """
        super().update_forward_refs(**localns)
        cls.__name__ = remove_forward_ref_notation(cls.__name__)
        cls.__qualname__ = remove_forward_ref_notation(cls.__qualname__)

    def validate_contents(self) -> None:
        self._validate_and_set_contents(self.contents)

    def _validate_and_set_contents(self, new_contents: RootT) -> None:
        self.contents = self._validate_contents_from_value(new_contents)
        self._take_snapshot_of_validated_contents()

    def _validate_contents_from_value(self, value: RootT) -> RootT:
        if isinstance(value, Model) and not is_none_type(value):
            value = value.to_data()
        values, fields_set, validation_error = validate_model(self.__class__, {ROOT_KEY: value})
        if validation_error:
            raise validation_error
        return values[ROOT_KEY]

    def _get_restorable_contents(self):
        if not id(self) in _restorable_content_cache:
            _restorable_content_cache[id(self)] = RestorableContents()
        return _restorable_content_cache.get(id(self))

    def _take_snapshot_of_validated_contents(self):
        interactive_mode = self._is_interactive_mode()
        if interactive_mode:
            self._get_restorable_contents().take_snapshot(self.contents)

    def _is_interactive_mode(self):
        from omnipy import runtime
        interactive_mode = runtime.config.data.interactive_mode if runtime else True
        return interactive_mode

    @classmethod
    def _parse_data(cls, data: RootT) -> Any:
        return data

    @root_validator
    def _parse_root_object(cls, root_obj: dict[str, RootT]) -> Any:  # noqa
        assert ROOT_KEY in root_obj
        value = root_obj[ROOT_KEY]
        value = cls._parse_none_value_with_root_type_if_model(value)
        return {ROOT_KEY: cls._parse_data(value)}

    # Partial workaround of https://github.com/pydantic/pydantic/issues/3836, together with
    # _propagate_allow_none_from_model().  See series of relevant tests in test_model.py
    # starting with test_nested_model_classes_none_as_default().
    @classmethod
    def _parse_none_value_with_root_type_if_model(cls, value):
        root_field = cls.__fields__.get(ROOT_KEY)
        root_type = root_field.type_
        value = cls._parse_with_root_type_if_model(value, root_field, root_type)
        return value

    @classmethod
    def _parse_with_root_type_if_model(cls, value: Any, root_field: ModelField,
                                       root_type: Type) -> Any:
        if get_origin(root_type) is Annotated:
            root_type = get_args(root_type)[0]

        if get_origin(root_type) is Union:
            last_error_holder = LastErrorHolder()
            for arg in get_args(root_type):
                with last_error_holder:
                    return cls._parse_with_root_type_if_model(value, root_field, arg)
            last_error_holder.raise_derived(NoneIsNotAllowedError())

        if lenient_issubclass(root_type, Model):
            if root_field.outer_type_ != root_type:
                outer_type = get_origin(root_field.outer_type_)
                if lenient_issubclass(outer_type, Sequence) and lenient_isinstance(value, Sequence):
                    return outer_type(
                        root_type.parse_obj(val)
                        if is_none_type(val) or not lenient_isinstance(val, Model) else val
                        for val in value)
                elif lenient_issubclass(outer_type, Mapping) and lenient_isinstance(value, Mapping):
                    return outer_type({
                        key:
                            root_type.parse_obj(val)
                            if is_none_type(val) or not lenient_isinstance(val, Model) else val
                        for (key, val) in value.items()
                    })
            else:
                return root_type.parse_obj(
                    value) if is_none_type(value) or not lenient_isinstance(value, Model) else value
        if value is None:
            none_default = root_field.default_factory() is None if root_field.default_factory \
                else root_field.default is None
            root_type_is_none = is_none_type(root_type)
            root_type_is_optional = get_origin(root_type) is Union \
                and any(is_none_type(arg) for arg in get_args(root_type))
            supports_none = none_default or root_type_is_none or root_type_is_optional
            if not supports_none:
                raise NoneIsNotAllowedError()

        return value

    @property
    def contents(self) -> RootT:
        return self.__dict__.get(ROOT_KEY)

    @contents.setter
    def contents(self, value: RootT) -> None:
        super().__setattr__(ROOT_KEY, value)

    def to_data(self) -> Any:
        return self.dict()[ROOT_KEY]

    def from_data(self, value: object) -> None:
        self._validate_and_set_contents(value)

    def absorb_and_replace(self, other: 'Model'):
        self.from_data(other.to_data())

    def to_json(self, pretty=True) -> str:
        json_content = self.json()
        if pretty:
            return self._pretty_print_json(json.loads(json_content))
        else:
            return json_content

    def from_json(self, json_contents: str) -> None:
        new_model = self.parse_raw(json_contents, proto=PydanticProtocol.json)
        self.contents = new_model.contents

    def inner_type(self, with_args: bool = False) -> type | None:
        return self.__class__._get_root_type(outer=False, with_args=with_args)

    def outer_type(self, with_args: bool = False) -> type | None:
        return self.__class__._get_root_type(outer=True, with_args=with_args)

    def is_nested_type(self) -> bool:
        return not self.inner_type(with_args=True) == self.outer_type(with_args=True)

    @classmethod
    def _get_root_type(cls, outer: bool, with_args: bool) -> type | None:
        root_field = cast(ModelField, cls.__fields__.get(ROOT_KEY))
        root_type = root_field.outer_type_ if outer else root_field.type_
        root_type = remove_annotated_plus_optional_if_present(root_type)
        return root_type if with_args else ensure_plain_type(root_type)

    # @classmethod
    # def create_from_json(cls, data: str | tuple[str]):
    #     if isinstance(data, tuple):
    #         data = data[0]
    #
    #     obj = cls()
    #     obj.from_json(data)
    #     return obj
    #
    # def __reduce__(self):
    #     return self.__class__.create_from_json, (self.to_json(),)

    @classmethod
    def to_json_schema(cls, pretty=True) -> str:
        schema = cls.schema()

        if pretty:
            return cls._pretty_print_json(schema)
        else:
            return json.dumps(schema)

    @staticmethod
    def _pretty_print_json(json_content: Any) -> str:
        return json.dumps(json_content, indent=2)

    def _check_for_root_key(self) -> None:
        if ROOT_KEY not in self.__dict__:
            raise TypeError('The Model class requires the specific model to be specified in as '
                            'a type hierarchy within brackets either directly, e.g.:\n'
                            '\t"model = Model[list[int]]([1,2,3])"\n'
                            'or indirectly in a subclass definition, e.g.:\n'
                            '\t"class MyNumberList(Model[list[int]]): ..."')

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in ['__module__'] + list(self.__dict__.keys()) and attr not in [ROOT_KEY]:
            super().__setattr__(attr, value)
        else:
            if attr in ['contents']:
                contents_prop = getattr(self.__class__, attr)
                contents_prop.__set__(self, value)
            else:
                raise RuntimeError('Model does not allow setting of extra attributes')

    def _special_method(self, name: str, info: MethodInfo, *args: object,
                        **kwargs: object) -> object:
        method = self._getattr_from_contents(name)

        if info.state_changing:
            restorable = self._get_restorable_contents()
            if self._is_interactive_mode():
                if restorable.has_snapshot() \
                        and restorable.last_snapshot_taken_of_same_obj(self.contents) \
                        and restorable.differs_from_last_snapshot(self.contents):

                    # Current contents not validated
                    reset_contents_to_last_snapshot = AttribHolder(
                        self, 'contents', restorable.get_last_snapshot(), reset_to_other=True)
                    with reset_contents_to_last_snapshot:
                        validated_contents = self._validate_contents_from_value(self.contents)

                    reset_contents_to_validated_prev = AttribHolder(
                        self, 'contents', validated_contents, reset_to_other=True)
                    reset_solution = reset_contents_to_validated_prev
                else:
                    reset_contents_to_prev = AttribHolder(self, 'contents', copy_attr=True)
                    reset_solution = reset_contents_to_prev
            else:
                reset_solution = nothing()

            with reset_solution:
                ret = method(*args, **kwargs)
                if self._is_interactive_mode():
                    needs_validation = restorable.differs_from_last_snapshot(self.contents) \
                        if restorable.has_snapshot() else True
                else:
                    needs_validation = True
                if needs_validation:
                    self.validate_contents()
        else:
            ret = method(*args, **kwargs)

        if info.maybe_returns_same_type:
            ret = self._convert_to_model_if_reasonable(args, name, ret)

        return ret

    def _convert_to_model_if_reasonable(self, args, name, ret):
        if not isinstance(ret, self.__class__):
            if name == '__getitem__':
                assert len(args) == 1
                if not isinstance(args[0], slice):
                    # assert self.is_nested_type()
                    # TODO: With Python 3.13 and PEP 649, reconsider the choice to not automatically
                    #       generate nested models through '__getitem__'.
                    #
                    # Seems to work, but the consequences for this are big and requires thought
                    # and consideration.
                    #
                    # try:
                    #     ret = Model[self.inner_type(with_args=True)](ret)
                    # except ValidationError:
                    #     pass
                    return ret

            # We can do this with some ease of mind as all the methods except '__getitem__' with
            # integer argument are supposed to possibly return a result of the same type.
            outer_type = self.outer_type(with_args=True)
            outer_type_plain = self.outer_type()

            types_to_check = get_args(outer_type) if is_union(outer_type) else [outer_type_plain]
            for type_to_check in types_to_check:
                if type_to_check is not None and isinstance(ret, ensure_plain_type(type_to_check)):
                    try:
                        ret = self.__class__(ret)
                    except ValidationError:
                        pass
        return ret

    def __getattr__(self, attr: str) -> Any:
        ret = self._getattr_from_contents(attr)
        if callable(ret):
            contents_holder = AttribHolder(self, 'contents', copy_attr=True)
            context = contents_holder if self._is_interactive_mode() else nothing()
            ret = add_callback_after_call(ret, self.validate_contents, with_context=context)
        return ret

    def _getattr_from_contents(self, attr):
        if isinstance(self.contents, Model):
            return getattr(self.contents.contents, attr)
        else:
            return getattr(self.contents, attr)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Model) \
            and self.__class__ == other.__class__ \
            and all_equals(self.contents, other.contents) \
            and self.to_data() == other.to_data()  # last is probably unnecessary, but just in case

    def __repr__(self) -> str:
        return super().__repr__()

    def __repr_args__(self):
        return [(None, self.contents)]



from abc import ABC, abstractmethod
from io import BytesIO
import os
import tarfile
from tarfile import TarInfo
from typing import Any, Callable, IO, Type

from pydantic import ValidationError

from omnipy.api.protocols.private.log import CanLog
from omnipy.api.protocols.public.data import IsDataset, IsSerializer, IsTarFileSerializer


class Serializer(ABC):
    """"""
    @classmethod
    @abstractmethod
    def is_dataset_directly_supported(cls, dataset: IsDataset) -> bool:
        pass

    @classmethod
    @abstractmethod
    def get_dataset_cls_for_new(cls) -> Type[IsDataset]:
        pass

    @classmethod
    @abstractmethod
    def get_output_file_suffix(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def serialize(cls, dataset: IsDataset) -> bytes | memoryview:
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized: bytes) -> IsDataset:
        pass


class TarFileSerializer(Serializer, ABC):
    """"""
    @classmethod
    def create_tarfile_from_dataset(cls,
                                    dataset: IsDataset,
                                    data_encode_func: Callable[[Any], bytes | memoryview]):
        bytes_io = BytesIO()
        with tarfile.open(fileobj=bytes_io, mode='w:gz') as tarfile_stream:
            for data_file, data in dataset.items():
                json_data_bytestream = BytesIO(data_encode_func(data))
                json_data_bytestream.seek(0)
                tarinfo = TarInfo(name=f'{data_file}.{cls.get_output_file_suffix()}')
                tarinfo.size = len(json_data_bytestream.getbuffer())
                tarfile_stream.addfile(tarinfo, json_data_bytestream)
        return bytes_io.getbuffer().tobytes()

    @classmethod
    def create_dataset_from_tarfile(cls,
                                    dataset: IsDataset,
                                    tarfile_bytes: bytes,
                                    data_decode_func: Callable[[IO[bytes]], Any],
                                    dictify_object_func: Callable[[str, Any], dict | str],
                                    import_method='from_data'):
        with tarfile.open(fileobj=BytesIO(tarfile_bytes), mode='r:gz') as tarfile_stream:
            for filename in tarfile_stream.getnames():
                data_file = tarfile_stream.extractfile(filename)
                assert filename.endswith(f'.{cls.get_output_file_suffix()}')
                data_file_name = '.'.join(filename.split('.')[:-1])
                getattr(dataset, import_method)(
                    dictify_object_func(data_file_name, data_decode_func(data_file)))


class SerializerRegistry:
    def __init__(self) -> None:
        self._serializer_classes: list[Type[IsSerializer]] = []

    def register(self, serializer_cls: Type[IsSerializer]) -> None:
        self._serializer_classes.append(serializer_cls)

    @property
    def serializers(self) -> tuple[Type[IsSerializer], ...]:
        return tuple(self._serializer_classes)

    @property
    def tar_file_serializers(self) -> tuple[Type[IsTarFileSerializer], ...]:
        return tuple(cls for cls in self._serializer_classes if issubclass(cls, TarFileSerializer))

    def auto_detect(self, dataset: IsDataset):
        return self._autodetect_serializer(dataset, self.serializers)

    def auto_detect_tar_file_serializer(self, dataset: IsDataset):
        return self._autodetect_serializer(dataset, self.tar_file_serializers)

    @classmethod
    def _autodetect_serializer(cls, dataset, serializers):
        # def _direct(dataset: Dataset, serializer: Serializer):
        #     new_dataset_cls = serializer.get_dataset_cls_for_new()
        #     new_dataset = new_dataset_cls(dataset)
        #     return new_dataset

        def _to_data_from_json(dataset: IsDataset, serializer: IsSerializer):
            new_dataset_cls = serializer.get_dataset_cls_for_new()
            new_dataset = new_dataset_cls()
            new_dataset.from_json(dataset.to_data())
            return new_dataset

        def _to_data_from_data(dataset: IsDataset, serializer: IsSerializer):
            new_dataset_cls = serializer.get_dataset_cls_for_new()
            new_dataset = new_dataset_cls()
            new_dataset.from_data(dataset.to_data())
            return new_dataset

        def _to_data_from_data_if_direct(dataset: IsDataset, serializer: IsSerializer):
            assert serializer.is_dataset_directly_supported(dataset)
            return _to_data_from_data(dataset, serializer)

        # def _to_json_from_json(dataset: Dataset, serializer: Serializer):
        #     new_dataset_cls = serializer.get_dataset_cls_for_new()
        #     new_dataset = new_dataset_cls()
        #     new_dataset.from_json(dataset.to_json())
        #     return new_dataset

        for func in (_to_data_from_data_if_direct, _to_data_from_json, _to_data_from_data):
            for serializer in serializers:
                try:
                    new_dataset = func(dataset, serializer)
                    return new_dataset, serializer
                except (TypeError, ValueError, ValidationError, AssertionError):
                    pass

        return None, None

    def detect_tar_file_serializers_from_file_suffix(self, file_suffix: str):
        return tuple(serializer_cls for serializer_cls in self.tar_file_serializers
                     if serializer_cls.get_output_file_suffix() == file_suffix)

    def load_from_tar_file_path(self, log_obj: CanLog, tar_file_path: str, to_dataset: IsDataset):
        if hasattr(log_obj, 'log'):
            log = log_obj.log
        else:
            log = print

        with tarfile.open(tar_file_path, 'r:gz') as tarfile_obj:
            file_suffixes = set(fn.split('.')[-1] for fn in tarfile_obj.getnames())
        if len(file_suffixes) != 1:
            log(f'Tar archive contains files with different or '
                f'no file suffixes: {file_suffixes}. Serializer '
                f'cannot be uniquely determined. Aborting '
                f'restore.')
        else:
            file_suffix = file_suffixes.pop()
            serializers = self.detect_tar_file_serializers_from_file_suffix(file_suffix)
            if len(serializers) == 0:
                log(f'No serializer for file suffix "{file_suffix}" can be'
                    f'determined. Aborting restore.')
            else:
                log(f'Reading dataset from a gzipped tarpack at'
                    f' "{os.path.abspath(tar_file_path)}"')

                serializer = serializers[0]
                with open(tar_file_path, 'rb') as tarfile_binary:
                    auto_dataset = serializer.deserialize(tarfile_binary.read())

                if to_dataset.get_model_class() is auto_dataset.get_model_class():
                    to_dataset.data = auto_dataset.data
                    return to_dataset
                else:
                    try:
                        if to_dataset.get_model_class().inner_type == str:
                            to_dataset.from_data(auto_dataset.to_json())
                        else:
                            to_dataset.from_json(auto_dataset.to_data())
                        return to_dataset
                    except Exception:
                        return auto_dataset



from typing import Callable, TypeAlias, TypeVar

GeneralDecorator = Callable[[Callable], Callable]
LocaleType: TypeAlias = str | tuple[str | None, str | None]

DecoratorClassT = TypeVar('DecoratorClassT', covariant=True)
JobConfigT = TypeVar('JobConfigT', covariant=True)
JobT = TypeVar('JobT', covariant=True)
JobTemplateT = TypeVar('JobTemplateT', covariant=True)
TaskTemplateT = TypeVar('TaskTemplateT')
TaskTemplateContraT = TypeVar('TaskTemplateContraT', contravariant=True)
TaskTemplateCovT = TypeVar('TaskTemplateCovT', covariant=True)



from typing import Protocol

from omnipy.api.enums import (ConfigOutputStorageProtocolOptions,
                              ConfigPersistOutputsOptions,
                              ConfigRestoreOutputsOptions)
from omnipy.api.typedefs import LocaleType


class IsEngineConfig(Protocol):
    """"""
    ...


class IsLocalRunnerConfig(IsEngineConfig, Protocol):
    """"""
    ...


class IsPrefectEngineConfig(IsEngineConfig, Protocol):
    """"""
    use_cached_results: bool = False


class IsJobConfig(Protocol):
    """"""
    output_storage: 'IsOutputStorage'


class IsDataConfig(Protocol):
    """"""
    interactive_mode: bool


class IsRootLogConfig(Protocol):
    """"""
    log_format_str: str
    locale: LocaleType
    log_to_stdout: bool
    log_to_stderr: bool
    log_to_file: bool
    stdout_log_min_level: int
    stderr_log_min_level: int
    file_log_min_level: int
    file_log_dir_path: str


class IsOutputStorageBase(Protocol):
    persist_data_dir_path: str


class IsLocalOutputStorage(IsOutputStorageBase, Protocol):
    ...


class IsS3OutputStorage(IsOutputStorageBase, Protocol):
    endpoint_url: str
    access_key: str
    secret_key: str
    bucket_name: str


class IsOutputStorage(Protocol):
    persist_outputs: ConfigPersistOutputsOptions
    restore_outputs: ConfigRestoreOutputsOptions
    protocol: ConfigOutputStorageProtocolOptions
    local: IsLocalOutputStorage
    s3: IsS3OutputStorage



from contextlib import AbstractContextManager, contextmanager
from copy import deepcopy

from omnipy.util.helpers import all_equals

# TODO: Consider refactoring as many as possible of the context managers (AbstractContextManager
#       subclasses) to @contextmanager-decorated methods


class LastErrorHolder(AbstractContextManager):
    def __init__(self):
        self._last_error = None

    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            self._last_error = exc_val
        return True

    def raise_derived(self, exc: Exception):
        if self._last_error is not None:
            raise exc from self._last_error
        else:
            raise exc


Undefined = object()


# TODO: Perhaps the two use cases of this are so dissimilar that the class should be split into two
#       distinct subclasses to make the code easier to understand?
class AttribHolder(AbstractContextManager):
    def __init__(self,
                 obj: object,
                 attr_name: str,
                 other_value: object = Undefined,
                 reset_to_other: bool = False,
                 switch_to_other: bool = False,
                 on_class: bool = False,
                 copy_attr: bool = False):
        self._obj_or_cls = obj.__class__ if on_class else obj
        self._attr_name = attr_name
        self._other_value = None if other_value is Undefined else other_value
        self._prev_value: object | None = None
        self._reset_to_other = reset_to_other
        self._switch_to_other = switch_to_other
        self._copy_attr = copy_attr
        self._store_prev_attr = False
        self._set_attr_to_other = False

        assert not (reset_to_other and switch_to_other), \
            'Only one of `reset_to_other` and `switch_to_other` can be specified.'

        if other_value is not Undefined:
            assert reset_to_other or switch_to_other, \
                ('If other_value is specified, you also need to set one of `reset_to_other` and '
                 '`switch_to_other`')

        if reset_to_other or switch_to_other:
            assert other_value is not Undefined, \
                ('If one of `reset_to_other` and `switch_to_other` are specified, you also need '
                 'to provide a value for `other_value`')

    def __enter__(self):
        if hasattr(self._obj_or_cls, self._attr_name):
            attr_value = getattr(self._obj_or_cls, self._attr_name)

            self._store_prev_attr = \
                not self._reset_to_other and not all_equals(attr_value, self._other_value)

            if self._store_prev_attr:
                self._prev_value = deepcopy(attr_value) if self._copy_attr else attr_value

            if self._switch_to_other:
                setattr(self._obj_or_cls, self._attr_name, self._other_value)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._switch_to_other or exc_val is not None:
            reset_value = self._other_value if self._reset_to_other else self._prev_value
            setattr(self._obj_or_cls, self._attr_name, reset_value)

        if self._store_prev_attr:
            self._prev_value = None


@contextmanager
def nothing(*args, **kwds):
    yield None


class PrintExceptionContext(AbstractContextManager):
    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f'{exc_type.__name__}: {str(exc_val).splitlines()[0]}', end='')
        return True


print_exception = PrintExceptionContext()



from omnipy.compute.flow import FuncFlowTemplate
from omnipy.compute.typing import mypy_fix_func_flow_template

from .datasets import JsonListOfDictsDataset, JsonListOfDictsOfScalarsDataset
from .tasks import DEFAULT_KEY, flatten_outer_level_of_all_data_files, ID_KEY, REF_KEY


@mypy_fix_func_flow_template
@FuncFlowTemplate(fixed_params=dict(id_key=ID_KEY, ref_key=REF_KEY, default_key=DEFAULT_KEY))
def flatten_nested_json(
    dataset: JsonListOfDictsDataset,
    id_key: str,
    ref_key: str,
    default_key: str,
) -> JsonListOfDictsOfScalarsDataset:
    all_data_files_of_scalar_records_ds = JsonListOfDictsOfScalarsDataset()
    data_files_of_any_ds = dataset

    while len(data_files_of_any_ds) > 0:
        data_files_of_scalar_records_ds, data_files_of_any_ds = \
            flatten_outer_level_of_all_data_files(
                data_files_of_any_ds, id_key, ref_key, default_key)

        all_data_files_of_scalar_records_ds |= data_files_of_scalar_records_ds

    return all_data_files_of_scalar_records_ds



from typing import Generic, Hashable, TypeAlias, TypeVar

from omnipy.data.model import Model
from omnipy.modules.general.typedefs import FrozenDict
from omnipy.util.helpers import is_iterable


class NotIterableExceptStrOrBytesModel(Model[object | None]):
    """
    Model describing any object that is not iterable, except for `str` and `bytes` types.
    As strings and bytes are iterable (over the characters/bytes) but also generally useful and
    often considered singular (or scalar) types, they are specifically allowed by this model.

    Examples:
        >>> from omnipy import NotIterableExceptStrOrBytesModel, print_exception
        >>>
        >>> NotIterableExceptStrOrBytesModel(1234)
        NotIterableExceptStrOrBytesModel(1234)
        >>> NotIterableExceptStrOrBytesModel('1234')
        NotIterableExceptStrOrBytesModel(1234)
        >>> with print_exception:
        ...     NotIterableExceptStrOrBytesModel((1, 2, 3, 4))
        ValidationError: 1 validation error for NotIterableExceptStrOrBytesModel

    Note:
        JsonScalarModel is a strict submodel of NotIterableExceptStrOrBytesModel in that all objects
        allowed by JsonScalarModel are also allowed by NotIterableExceptStrOrBytesModel.
    """
    @classmethod
    def _parse_data(cls, data: object) -> object:
        assert isinstance(data, str) or isinstance(data, bytes) or not is_iterable(data), \
            f'Data of type {type(data)} is iterable'
        return data


# TODO: Follow pydantic topic https://github.com/pydantic/pydantic/issues/6868 on MappingProxyType.
#       Used way too much energy to implement (and test) recursive frozen models, only to discover
#       that pydantic actually does not support MappingProxyType, which was hugely surprising.
#       Add tests on the immutability part once there is support for this.

# Note: see comments in omnipy.modules.json.models for more information on the basic organisation of
#       nested models.

#
# Private models
#

# Basic building block models

_KeyT = TypeVar('_KeyT', bound=str | Hashable)
_ValT = TypeVar('_ValT', bound=NotIterableExceptStrOrBytesModel | object)

_FrozenBaseT = TypeVar('_FrozenBaseT')

# class _FrozenScalarM(NotIterableExceptStrOrBytesModel):
#     ...


class _FrozenScalarM(Model[_ValT], Generic[_ValT]):
    _parse_data = NotIterableExceptStrOrBytesModel._parse_data


class _FrozenTupleBaseM(Model[tuple[_FrozenBaseT, ...]], Generic[_FrozenBaseT]):
    ...


class _FrozenDictBaseM(Model[FrozenDict[_KeyT, _FrozenBaseT]], Generic[_KeyT, _FrozenBaseT]):
    ...


class _FrozenTupleM(_FrozenTupleBaseM['_FrozenAnyUnion'], Generic[_ValT]):
    ...


#
# class _FrozenDictM(_FrozenDictBaseM[_KeyT, '_FrozenAnyUnion'], Generic[_KeyT, _ValT]):
#     ...


class _FrozenDictM(_FrozenDictBaseM[str | Hashable, '_FrozenAnyUnion'], Generic[_KeyT, _ValT]):
    ...


class _FrozenNoDictsM(_FrozenTupleBaseM['_FrozenNoDictsUnion'], Generic[_ValT]):
    ...


# class _FrozenNoTuplesM(_FrozenDictBaseM['_KeyT', '_FrozenNoTuplesUnion'], Generic[_KeyT, _ValT]):
#     ...


class _FrozenNoTuplesM(_FrozenDictBaseM[str | Hashable, '_FrozenNoTuplesUnion'],
                       Generic[_KeyT, _ValT]):
    ...


# TypeAliases

_FrozenAnyUnion: TypeAlias = \
    _FrozenScalarM[_ValT] | _FrozenTupleM[_ValT] | _FrozenDictM[_KeyT, _ValT]
_FrozenNoDictsUnion: TypeAlias = _FrozenScalarM[_ValT] | _FrozenNoDictsM[_ValT]
_FrozenNoTuplesUnion: TypeAlias = _FrozenScalarM[_ValT] | _FrozenNoTuplesM[_KeyT, _ValT]

# Basic models needs to update their forward_refs with type aliases declared above

_FrozenTupleM.update_forward_refs()
_FrozenDictM.update_forward_refs()
_FrozenNoDictsM.update_forward_refs()
_FrozenNoTuplesM.update_forward_refs()

#
# Exportable models
#

# General


class NestedFrozenDictsOrTuplesModel(Model[_FrozenAnyUnion], Generic[_KeyT, _ValT]):
    """
    Recursive model for nested immutable containers (FrozenDict and tuples). Not functional.

    Awaiting support for MappingProxyType in pydantic for the immutability to actually work.
    Also use as generic types awaits Python 3.13. Way ahead of the crowd here!
    """


class NestedFrozenTuplesModel(Model[_FrozenNoDictsM[_ValT]], Generic[_ValT]):
    """
    Recursive model for nested tuples.

    Awaiting support for MappingProxyType in pydantic for the immutability to actually work.
    Also use as generic types awaits Python 3.13. Way ahead of the crowd here!
    """


class NestedFrozenDictsModel(Model[_FrozenNoTuplesM[_KeyT, _ValT]], Generic[_KeyT, _ValT]):
    """
    Recursive model for nested FrozenDicts.

    Awaiting support for MappingProxyType in pydantic for the immutability to actually work.
    Also use as generic types awaits Python 3.13. Way ahead of the crowd here!
    """
